package Report;

use strict;
use warnings;

=head1 Combined.pm (Report)

Complex report - handles combined data with various keys grouping and
aggregation functions.

=head2 Generates

Generates 1D, 2D or 3D report, depending on selections in config.

In 1D reports tables contain keyfields in rows and aggregations
in columns. Single table for each grouping.

In 2D reports tables contain keyfields in rows and columns.
Separate tables for each aggregation.

In 3D reports 2D tables are generated for each keyvalue of 3rd group.
That is - additional group defines 3rd dimention, each 'row' of which
goes in separate table.
Separate tables for each aggregation and keyvalue.

=head2 Config

Accepts usual selections to add single keyfield and aggregation:

    select: fookey;
    select: fun(foo);

Single grouping produces rowspaned 1D tables:

    select: GROUP;

Double grouping produces joined 2D tables:

    select: ROWGROUP / COLGROUP;

Triple grouping makes defines 2D tables for each value of keys in 3rd
group:

    select: ROWGROUP / COLGROUP / SPLITGROUP;

There, of cos, may be several selections in several directives.

=head2 Remarks

TODO: Some config-option to setup whether to split tables by aggregations
      or put them all in same large table in wallpaper format.
      Another config-option to setup whether print all tables on same page
      or on alot of separate pages.
      NOTE: They should affect sub generate() only.

=head2 Debugging

    ::DEBUG bit 1 set (==2) - dumps loaded data and exits;
    ::DEBUG bit 4 set (==16) - verbose initialization (configtest);

=cut

use vars qw(@pages_1d @pages_2d @pages_3d $VERSION);
use Exporter;
use Core
qw#
	@keyfields @selections @aggregations %aggregations @sortorder
	%datatitles $reportype_title @pages $data_ref $table_refname
	$table_caption

	keysort unroll setup format_titles strfval file_idx_ref
	generate_table generate_general generate_head
#;
use Utils qw#cmp_arrays printgrouping trans#;
use Html qw#foot escape#;

@Report::EXPORT_OK = qw#generate page_title file_ref page_header#;

@Report::ISA = ('Exporter');

$Report::NAME = 'combined';
$Report::VERSION = '0.1.0';
$Report::AUTHOR = 'qMax';
$Report::EMAIL = 'qmax@mail.ru';

# list of groups combinations
# all groups are stored by reference, because reused
# but some groups may be the same

# 3d: [ \@rowgroup \@colgroup \@splitgroup ]
@pages_3d=();
# 2d: [ \@rowgroup \@colgroup undef ]
@pages_2d=();
# 1d: [ \@rowgroup undef      undef ]
@pages_1d=();

print "\nCombined.pm (Report) loaded. Calling 'setup ()'." if ($::DEBUG & 0x01);

#### BEGIN

# imported from Core
setup ();

unless(@keyfields) {
    print STDERR "\nreport/$Report::NAME: FATAL: no data to report defined";
    die "\nCheck your config file\n";
    }

# NB: it's OK to dublicate indexes,
#     data module will filter them out
my @indexes=();
foreach my $groupset ( @selections ) {

    my $rowgroups = shift @$groupset;
    my $colgroups = shift @$groupset;
    my $splitgroups = shift @$groupset;

    if( $splitgroups ) {
	# 3D
	foreach my $gr ( @$splitgroups ) {
	    foreach my $rowgr ( @$rowgroups ) {
		foreach my $colgr ( @$colgroups ) {
		    next if( grep ( (cmp_arrays($gr,$_->[2])
				     and cmp_arrays($rowgr,$_->[0])
				     and cmp_arrays($colgr,$_->[1])),
				    @pages_3d)
			     );
		    push @pages_3d, [ $rowgr, $colgr, $gr];
		    # all recombinations of indexes
		    push @indexes, [@$rowgr, @$colgr, @$gr];
		    # to get totals for rows and columns
		    push @indexes, [@$rowgr, @$gr];
		    push @indexes, [@$colgr, @$gr];
		    # to get total
		    push @indexes, [@$gr];
		    }
		}
	    }
	}
    elsif( $colgroups ) {
	# 2D
	foreach my $rowgr ( @$rowgroups ) {
	    foreach my $colgr ( @$colgroups ) {
		next if( grep ( (cmp_arrays($rowgr,$_->[0])
				 and cmp_arrays($colgr,$_->[1])),
				@pages_2d)
			 );
		push @pages_2d, [$rowgr, $colgr, undef];
		push @indexes, [@$rowgr, @$colgr];
		# to get all totals for rows and cols
		push @indexes, [@$rowgr];
		push @indexes, [@$colgr];
		}
	    }
	}
    else {
	# 1D
	foreach my $rowgr ( @$rowgroups ) {
	    next if( grep (cmp_arrays($rowgr,$_->[0]), @pages_1d) );
	    push @pages_1d, [$rowgr, undef, undef ];
	    push @indexes,  [@$rowgr];
	    }
	}
    }

@pages=(@pages_1d,@pages_2d,@pages_3d);

my $have_pages = ( @pages_3d ) ? "3D" :
              ( @pages_2d ) ? "2D" :
			      "rowspaned";

$reportype_title="$have_pages";
print STDERR "\nreport/$Report::NAME: Generating combined report with $have_pages tables." if( $::DEBUG );

if( $::DEBUG&16 ) {
    print "\nreport/$Report::NAME: Using data fields:",
	"\n\t(".join(" ", @keyfields).")";

    print "\nreport/$Report::NAME: Using grouping:";
    foreach my $gr (@pages) {
	print "\n\t".printgrouping(@$gr);
	}

    print "\nreport/$Report::NAME: Using aggregations:";
    print "\n\t".join(" ",@aggregations);
    }

Data::setup([ [[],@indexes], [values %aggregations] ]);

#### END of BEGIN

sub generate
{
    my ($directory,$filename);
    if ($::DEBUG&2) { Data::dump(); return; }

    format_titles();

    #
    # I want all files to be created in single place.
    # And i ever know - why...
    #
    $directory=$main::outputdir;

    $filename=$directory."index.html";
    print STDERR "\nGeneral page..." if $::DEBUG;
    open (HTMLFILE, "> $filename") or die "when opening $filename: $!";
    generate_general(*HTMLFILE);
    close(HTMLFILE);

    foreach my $category (@pages_1d) {
	my($rowgr,$colgr,$splgr)=@$category;
	$filename = $directory.file_ref(@$category);
	print STDERR "\n1D page in ".file_ref(@$category)."..." if $::DEBUG;
	open (HTMLFILE, "> $filename") or die "$!\n";
	generate_page1D(*HTMLFILE,@$category);
	print HTMLFILE foot();
	close(HTMLFILE);
	}

    foreach my $category (@pages_2d) {
	my($rowgr,$colgr,$splgr)=@$category;
	$filename=$directory.file_ref(@$category);
	print STDERR "\n2D page in ".file_ref(@$category)."..." if $::DEBUG;
	open (HTMLFILE, "> $filename") or die "$!\n";
	generate_page2D(*HTMLFILE,@$category);
	print HTMLFILE foot();
	close(HTMLFILE);
	}

    foreach my $category (@pages_3d) {
	my($rowgr,$colgr,$splgr)=@$category;
	$filename=$directory.file_ref(@$category);
	print STDERR "\n3D page in ".file_ref(@$category)."..." if $::DEBUG;
	open (HTMLFILE, "> $filename") or die "$!\n";
	generate_page3D(*HTMLFILE,@$category);
	print HTMLFILE foot();
	close(HTMLFILE);
	}
}

sub generate_page3D
{
    my ($FILE,@category)=@_;
    my ($rowfields,$colfields,$fields)=@category;

    @sortorder=@$fields;
    my @basekeys=sort keysort unroll($fields);

    generate_head($FILE,@category);
    print {$FILE} "\n<hr>";

    $data_ref=\&data_xref;
    generate_table($FILE,{},\@aggregations,\@basekeys,$fields);
    $data_ref=undef;

    $table_caption = \&table_title;
    $table_refname = \&table_name;

    foreach my $basekey ( @basekeys ) {
	
	@sortorder=@$rowfields;
	my @rowkeys=sort keysort unroll($rowfields,$basekey);
	@sortorder=@$colfields;
	my @colkeys=sort keysort unroll($colfields,$basekey);
	
	foreach my $aggr (@aggregations) {
	    generate_table($FILE,$basekey,[$aggr],\@rowkeys,$rowfields,\@colkeys,$colfields);
	    print {$FILE} '<br>';
	    }
	}
}

sub generate_page2D
{
    my ($FILE,@category)=@_;
    my ($rowfields,$colfields)=@category;

    @sortorder=@$rowfields;
    my @rowkeys=sort keysort unroll($rowfields);
    @sortorder=@$colfields;
    my @colkeys=sort keysort unroll($colfields);

    generate_head($FILE,@category);

    if( @aggregations > 1 ) {
	print {$FILE} "\n<table class='index'><tr><td class='index' valign=middle border=0>";
	foreach my $cat (@aggregations) {
	    my $link = '#'.table_name(@{$aggregations{$cat}});
	    my $title = $datatitles{$cat};
	    print {$FILE} "\n<a class='index' href='$link'>".
			  "[\&nbsp;$title\&nbsp;]".
			  "</a>";
	    }
	print {$FILE} "\n</table>";
	}

    print {$FILE} "\n<hr>";

    $table_refname = \&table_name;
    foreach my $aggr (@aggregations) {
	my @agg = @{ $aggregations{$aggr} };
	#flatkeys(\@rowkeys);
	#flatkeys(\@colkeys);
	generate_table($FILE,{},[$aggr],\@rowkeys,$rowfields,\@colkeys,$colfields);
	print {$FILE} '<br>';
	}
}

sub generate_page1D
{
    my ($FILE,@category)=@_;
    my ($rowfields)=@category;

    @sortorder=@$rowfields;
    my @rowkeys=sort keysort unroll($rowfields);

    generate_head($FILE,@category);
    print {$FILE} "\n<hr>";
    generate_table($FILE,{},\@aggregations,\@rowkeys,$rowfields);
}

sub file_ref
{
    my ($rg,$cg,$sg)=@_;
    my @h=();
    push @h,file_idx_ref(@$sg) if( $sg );
    push @h,file_idx_ref(@$rg) if( $rg );
    push @h,file_idx_ref(@$cg) if( $cg );
    # NB: should not be '-', as in idx_ref, since groups may overlap
    return join('.',@h).".html";
}

sub page_title
{
    # should be something more compact
    return page_header(@_);
}

sub page_header
{
    my ($rg,$cg,$sg)=@_;
    my @main = ();
    my @join = ();
    my $title;

    # report:   3D                       2D              1D
    # 3D table: @splited by @rows,@cols
    # 2D table:          by @rows,@cols  @rows by @cols
    # 1D table: @rows                    @rows           @rows
    if( $sg ) {
	@main = @$sg; @join = (@$rg,@$cg);
	}
    elsif( $cg ) {
	if( @pages_3d ) { @main = ();   @join = (@$rg,@$cg); }
	else            { @main = @$rg; @join = @$cg; }
	}
    else {
	@main = @$rg;
	}

    if( @main ) { $title .= trans('/%s/, /',@main); }
    if( @join ) {
	$title .= '&nbsp;' if( $title );
	$title .= trans('by /%s/, /', @join);
	}
    return $title;
}

sub table_title {
    my %key=@_;
    # grep out aggregation marks
    %key = map { $_=>$key{$_} } grep($_!~/[A-Z]/,keys %key);
    return trans('/%s:%s/, /',map( ($_,strfval($_,$key{$_})), keys %key));
}

sub table_name
{
    my %key=@_;
    return join('-',escape(map ("$_:$key{$_}",sort keys %key)));
}

sub data_xref
{
    my %key=@_;
    # link only from data cells
    return undef unless (grep($_=~/[A-Z]/,keys %key));
    return '#'.table_name(@_);
}


=head1 AUTHOR

QMax E<lt>qmax@mail.ruE<gt>

=cut

1;
